#-----------------------------------------------------------------------------
#
# DirectX .X, Ogre3d .mesh.xml and WaveFront .Obj File Exporter for Google SketchUp
# Author: Zbigniew Skowron, zbychs@gmail.com
#
# I'd like to thank Jonathan Harris and Erwan de Cadoudal for their exporters.
# They helped me a lot.
#
# Export to .obj based on "WaveFront .Obj File Exporter" by
# M van der Honing, mvdhoning@noeska.com
#
#-----------------------------------------------------------------------------

module ZbylsXExporter

module_function

# hacks that won't be necessary in Ruby 1.8.7 or above
def help_group_by(arr)
  res        = {}
  arr.each{|e| (res[yield(e)] ||= []) << e}
  res
end

def help_cluster_by(arr, &block)
 #group_by(&block).values        # In case of unsortable keys.
 help_group_by(arr, &block).sort.transpose.pop || []
end
   
Meter_to_inch = 39.3700787
Inch_to_meter = 1.0 / Meter_to_inch

FMT_X = 0
FMT_OGRE = 1
FMT_OBJ = 2

def fmtValFunc(fmtVal, valX, valOgre, valObj)
	return fmtVal == FMT_X ? valX : (fmtVal == FMT_OGRE ? valOgre : valObj);
end

#-----------------------------------------------------------------------------

def reload
	pfile = Sketchup.find_support_file "ZbylsXExporter.rb", "Plugins/ZbylsXExporter"
	load pfile
end

#-----------------------------------------------------------------------------

def directx_header()
	text = "xof 0303txt 0032\n"
end

def ogre_header()
	text = "<!-- Exported from SketchUp with ZbylsXExporter -->"
end

def ogre_header_mat()
	text = "// Exported from SketchUp with ZbylsXExporter"
end

def obj_header()
	text = "#Exported from SketchUp with ZbylsXExporter"
end

#-----------------------------------------------------------------------------

def directx_def_material()
text = "Material Default_Material { 
   1.0;1.0;1.0;1.0;;
   3.2;
   1.000000;1.000000;1.000000;;
   0.000000;0.000000;0.000000;;
} 
"
end

def ogre_def_material()
text = "material Default_Material
{
	technique
	{
		pass
		{
			ambient 1 1 1 1
			diffuse 1 1 1 1
			specular 1 1 1 1 3.2
		}
	}
} 
"
end

def obj_def_material()
text = "newmtl Default_Material 
   Ka 1.000000 1.000000 1.000000
   Kd 1.0 1.0 1.0 
   Tr 1.0
   Ks 1.000000 1.000000 1.000000
   Ns 3.2


"
end

#-----------------------------------------------------------------------------

def directx_material(name, color, textureFile, alpha)
# #{color.red / 255.0};#{color.green / 255.0};#{color.blue / 255.0};#{color.alpha / 255.0};;
	col = format("%8.4f;%8.4f;%8.4f;%8.4f;", 1.0, 1.0, 1.0, alpha / 255.0)
	if !textureFile
		col = format("%8.4f;%8.4f;%8.4f;%8.4f;",
				 color.red / 255.0, color.green / 255.0, color.blue / 255.0, alpha / 255.0)
	end
	specular = format("%8.4f", 3.2)
	tex = ""
	tex = "   TextureFilename { \"#{textureFile}\"; }\n" if textureFile
	
	text = "Material #{name} {
   #{col};
   #{specular};
   1.000000;1.000000;1.000000;;
   0.000000;0.000000;0.000000;;
#{tex}}
"
end

def fixTexureNameForObj(textureFile)
	return textureFile.delete(" ")
end

def obj_material(fixTexs, name, color, textureFile, alpha)
	col = format("%8.4f %8.4f %8.4f", 1.0, 1.0, 1.0)
	if !textureFile
		col = format("%8.4f %8.4f %8.4f",
				 color.red / 255.0, color.green / 255.0, color.blue / 255.0)
	end
	specular = format("%8.4f", 3.2)
	tex = ""
	mapKd = "map_Kd"
	mapKd = "mapKd" if fixTexs
	#texName = fixTexureNameForObj(textureFile) if textureFile
	texName = textureFile
	tex = "   #{mapKd} #{texName}" if textureFile
	
	text = "newmtl #{name}
   Ka 1.000000 1.000000 1.000000
   Kd #{col}
   Tr #{alpha/255.0}
   Ks 1.000000 1.000000 1.000000
   Ns #{specular}
   #{tex}

"
end

def ogre_material(name, color, textureFile, alpha)
	col = format("%8.4f %8.4f %8.4f", 1.0, 1.0, 1.0)
	if !textureFile
		col = format("%8.4f %8.4f %8.4f",
				 color.red / 255.0, color.green / 255.0, color.blue / 255.0)
	end
	specular = format("%8.4f", 3.2)
	tex = ""
	tex = "texture_unit
			{
				texture \"#{textureFile}\"
			}" if textureFile
	
	text = "material \"#{name}\"
{
	technique
	{
		pass
		{
			ambient 1 1 1 1
			diffuse #{col} #{alpha/255.0}
			specular 1 1 1 1 #{specular}

			#{tex}
		}
	}
}

"
end

#-----------------------------------------------------------------------------

def out_per_line_indent(f, str, indent = "  ")
	str.each_line { |line|
		f.puts("#{indent}#{line}")
	}
end

def out_tab(fmtVal, f, points, del = ",", extra = "", indent = "  ")
	cnt = points.size
	return if cnt < 1

	f.puts("  #{cnt};") if (fmtVal == FMT_X)

	for i in 0..(cnt - 2)
		p = points[i]
		out_per_line_indent(f,"#{p}#{del}", indent)
	end
		p = points[cnt - 1]
		out_per_line_indent(f,"#{p}#{extra};", indent) if (fmtVal == FMT_X)
		out_per_line_indent(f,"#{p}#{extra}", indent) if !(fmtVal == FMT_X)
end

#-----------------------------------------------------------------------------

def out_point(fmtVal, p, n, u, v)
	if (fmtVal != FMT_OGRE)
		return out_raw_point(fmtVal, p), out_normal(fmtVal, n), out_uv(fmtVal, u, v)
	end
	return out_ogre_point(fmtVal, p, n, u, v), "", ""
end

def out_ogre_point(fmtVal, p, n, u, v)
	return "<vertex>
	#{out_raw_point(fmtVal, p)}
	#{out_normal(fmtVal, n)}
	#{out_uv(fmtVal, u, v)}
</vertex>"
end

def out_raw_point(fmtVal, p)
	print "Kuozaa Too!\n" if (!p)
	return "nil" if (!p)
	px = p.x * Inch_to_meter
	py = p.z * Inch_to_meter
	pz = p.y * Inch_to_meter * (fmtVal == FMT_X ? 1 : -1)
	#text = "#{px.inspect};#{py.inspect};#{pz.inspect};"
	text = format("%8.4f;%8.4f;%8.4f;", px, py, pz) if (fmtVal == FMT_X)
	text = format("<position x=\"%.4f\" y=\"%.4f\" z=\"%.4f\" />", px, py, pz) if (fmtVal == FMT_OGRE)
	text = format("v %8.4f %8.4f %8.4f", px, py, pz) if (fmtVal == FMT_OBJ)
	text
end

#-----------------------------------------------------------------------------
 
def out_normal(fmtVal, p)
	print "Kuozaa Noo!\n" if (!p)
	return "nil" if (!p)
	px = p.x
	py = p.z
	pz = p.y * (fmtVal == FMT_X ? 1 : -1)
	#text = "#{p.x.inspect};#{p.y.inspect};#{p.z.inspect};"
	text = format("%8.4f;%8.4f;%8.4f;", px, py, pz) if (fmtVal == FMT_X)
	text = format("<normal x=\"%.4f\" y=\"%.4f\" z=\"%.4f\" />", px, py, pz) if (fmtVal == FMT_OGRE)
	text = format("vn %8.4f %8.4f %8.4f", px, py, pz) if (fmtVal == FMT_OBJ)
	text
end

#-----------------------------------------------------------------------------
 
def out_uv(fmtVal, u, v)
	#text = "#{u.inspect},#{v.inspect};"
	text = format("%8.4f,%8.4f;", u, v) if (fmtVal == FMT_X)
	text = format("<texcoord u=\"%.4f\" v=\"%.4f\" />", u, v) if (fmtVal == FMT_OGRE)
	text = format("vt %8.4f %8.4f", u, v) if (fmtVal == FMT_OBJ)
	text
end

#-----------------------------------------------------------------------------

def out_face(fmtVal, f, face, indent = "")
	face = face.reverse if (fmtVal == FMT_X)
	cnt = face.size
	#return if cnt < 3
	
	f.printf("  #{cnt};") if (fmtVal == FMT_X)
	f.printf("#{indent}<face ") if (fmtVal == FMT_OGRE)
	f.printf("f ") if (fmtVal == FMT_OBJ)
	for i in 0..(cnt - 2)
		p = face[i]
		f.printf("#{p[0]},") if (fmtVal == FMT_X)
		f.printf("v#{i+1}=\"#{p[0]}\" ") if (fmtVal == FMT_OGRE)
		f.printf("  #{p[0] + 1}/#{p[1] + 1}/#{p[2] + 1} ") if (fmtVal == FMT_OBJ)
	end
		p = face[cnt - 1]
		f.printf("#{p[0]}") if (fmtVal == FMT_X)
		f.printf("v#{cnt}=\"#{p[0]}\" />") if (fmtVal == FMT_OGRE)
		f.printf("  #{p[0] + 1}/#{p[1] + 1}/#{p[2] + 1}") if (fmtVal == FMT_OBJ)
end

#-----------------------------------------------------------------------------

# faces = [ face ]
# face = [ verts, mat ]
# verts = [ [pidx, uvidx, nidx], ... ]

def out_faces(fmtVal, f, faces, indent = "")
	cnt = faces.size
	return if cnt < 1
	
	f.puts("  #{cnt};") if (fmtVal == FMT_X)
	out_per_line_indent(f, "<faces count=\"#{cnt}\">", indent) if (fmtVal == FMT_OGRE)
	
	curMat = ""
	for i in 0..(cnt - 2)
		fc = faces[i]
		
		mat = fc[1]
		if (fmtVal == FMT_OBJ) && (mat != curMat)
			f.puts("usemtl #{mat}\n")
			curMat = mat
		end
		
		out_face(fmtVal, f, fc[0], indent + "	");
		f.puts(",") if (fmtVal == FMT_X)
		f.puts("\n") if !(fmtVal == FMT_X)
	end
		fc = faces[cnt - 1]

		mat = fc[1]
		if (fmtVal == FMT_OBJ) && (mat != curMat)
			f.puts("usemtl #{mat}\n")
			curMat = mat
		end

		out_face(fmtVal, f, fc[0], indent + "	");
		f.puts(";;") if (fmtVal == FMT_X)
		out_per_line_indent(f, "</faces>", "\n" + indent) if (fmtVal == FMT_OGRE)
end

#-----------------------------------------------------------------------------
 
def out_face_ns(fmtVal, f, face)
	return if !(fmtVal == FMT_X)
	
	face = face.reverse
	cnt = face.size
	#return if cnt < 3
	f.printf("  #{cnt};")
	for i in 0..(cnt - 2)
		p = face[i]
		f.printf("#{p[2]},")
	end
		p = face[cnt - 1]
		f.printf("#{p[2]}")
end
#-----------------------------------------------------------------------------
 
def out_normals(fmtVal, f, faces)
	return if !(fmtVal == FMT_X)

	cnt = faces.size
	return if cnt < 1
	f.puts("  #{cnt};")
	for i in 0..(cnt - 2)
		fc = faces[i]
		out_face_ns(fmtVal, f, fc[0]);
		f.puts(";")
	end
		fc = faces[cnt - 1]
		out_face_ns(fmtVal, f, fc[0]);
		f.puts(";;")
end

#-----------------------------------------------------------------------------
 
def out_face_materials(fmtVal, f, faces, materials)
	return if !(fmtVal == FMT_X)

	#mats = materials.keys #is this ok?
	mats = []
	for n, mat in materials
		mats.push(n)
	end
	f.puts("  #{mats.size};")
	cnt = faces.size
	return if cnt < 1
	f.puts("  #{cnt};")
	for i in 0..(cnt - 2)
		fc = faces[i]
		m = fc[1]
		midx = mats.index(m)
		f.puts("  #{midx},")
	end
		fc = faces[cnt - 1]
		m = fc[1]
		midx = mats.index(m)
		f.puts("  #{midx};")
		
	mats.each { |m|
		f.puts("   { #{m} }")
	}
end

#-----------------------------------------------------------------------------
 
def out_materials(fmtVal, fixTexs, f, materials)
	for n, mat in materials
		if n == "Default_Material"
			f.puts(directx_def_material()) if (fmtVal == FMT_X)
			f.puts(ogre_def_material()) if (fmtVal == FMT_OGRE)
			f.puts(obj_def_material()) if (fmtVal == FMT_OBJ)
		else
			f.puts(directx_material(mat[0], mat[1], mat[2], mat[3])) if (fmtVal == FMT_X)
			f.puts(ogre_material(mat[0], mat[1], mat[2], mat[3])) if (fmtVal == FMT_OGRE)
			f.puts(obj_material(fixTexs, mat[0], mat[1], mat[2], mat[3])) if (fmtVal == FMT_OBJ)
		end
	end
end

#-----------------------------------------------------------------------------

def processFileName(fname, fmtVal)
	slash = "\\"
	dir = "c:\\temp"
	sidx = fname.rindex(slash)
	if (!sidx)
		slash = "/"
		dir = "/tmp"
		sidx = fname.rindex(slash)
	end
	name = fmtValFunc(fmtVal, "Untitled.x", "Untitled.mesh.xml", "Untitled.obj");
	if (sidx)
		dir = fname[0..(sidx-1)]
		name = fname[(sidx+1)..-1]
		didx = name.rindex('.')
		ext = fmtValFunc(fmtVal, ".x", ".mesh.xml", ".obj")
		name += ext if !didx
		name[didx..-1] = ext if didx
	end
	
	return [name, dir, slash]
end

#-----------------------------------------------------------------------------

def exportXFileUI
	# Create the WebDialog instance

	fname = Sketchup.active_model.path
	name, dir, slash = processFileName(fname, FMT_OBJ)

	my_dialog = UI::WebDialog.new("Zbyl's .X and .Obj File Exporter", true, "Zbyl's .X and .Obj File Exporter", 
									600, 600, 200, 200, true)

	# Attach an action callback
	my_dialog.add_action_callback("browse") { |web_dialog, params|
		#puts params
		a = params.split(',');
  		outDir = a[0]
  		outFile = a[1]
  		fmtVal = Integer(a[2])
  		outDir += slash if outDir[-1] != slash[0]
		#puts(outDir, outFile)
		fname = UI.savepanel("Export to...", outDir, outFile)
		fname = UI.savepanel("Export to...") if !fname
		if fname
			name, dir, slash = processFileName(fname, fmtVal)
			dir = dir.gsub(/['"\\]/) { '\\'+$& }
  			my_dialog.execute_script(
				"document.getElementById('outDir').value = '#{dir}'");
  			my_dialog.execute_script(
				"document.getElementById('outFile').value = '#{name}'");
		end
	}

	my_dialog.add_action_callback("ruby") { |web_dialog, params|
		#puts params
		# display ruby panel for messages
		Sketchup.send_action "showRubyPanel:"
	}
	
	# Attach an action callback
	my_dialog.add_action_callback("export") { |web_dialog, params|
		#puts params
		a = params.split(',');
		#UI.messagebox("Ruby says: Your javascript has asked for " + a)
  		#my_dialog.execute_script(
  		#	"document.getElementById('ala').innerHTML= '#{a.to_s}'");
  		outDir = a[0]
  		outFile = a[1]
  		smartFaces = a[2] == "true" # guess which faces to export
  		front = a[3] == "true" # export front faces
  		back = a[4] == "true" # export back faces
  		uv = a[5] == "true" # use UVHelper
  		sel = a[6] == "true" #only selected
  		tex = a[7] == "true" # only textured
  		fmtVal = Integer(a[8]) # export as X File, Ogre or OBJ file
		fixTexs = a[9] == "true" # fix texs in OBJ
		fixUV = a[10] == "true" # alternative texture coord computation
		outName = outDir + slash + outFile

		#print(outName, outDir, outFile, smartFaces, front, back, uv, sel, tex)
  		my_dialog.execute_script(
			"document.getElementById('console').innerHTML = ''");
  		my_dialog.execute_script(
			"document.getElementById('progress').innerHTML = 'Exporting...'");
		
		res = exportXFile(outName, outDir, fmtVal, fixTexs, fixUV, smartFaces, front, back, uv, sel, tex,
		lambda {|forward, text|
			if forward
				puts text
				text = text.gsub(/['"\\]/) { '\\'+$& }
				text = text.gsub(/[\n]/) { '<br>' }
				text += '<br>'
		  		my_dialog.execute_script(
  					"document.getElementById('console').innerHTML += '\\n#{text}'");
			else
				print text
		  		#my_dialog.execute_script(
  				#	"document.getElementById('progress').innerHTML += '#{text}'");
			end
		})
  		my_dialog.execute_script(
			"document.getElementById('progress').innerHTML = '#{res ? 'Exported' : 'Not exported!'}'");
	}


	html="
	<html>
	<head>
		<title>WorldEd Exporter</title>
		<script>
		function callRuby(actionName, params) {
			query = 'skp:'+ actionName + '@' + params;
			window.location.href = query;
		}
		function rubyFunc() {
			callRuby('ruby', '')
		}
		function browseFunc() {
			outputDir = document.getElementById('outDir').value
			outputFile = document.getElementById('outFile').value
			fmtX = document.getElementById('asXFile').checked
			fmtOgre = document.getElementById('asOgreFile').checked
			fmtVal = fmtX ? 0 : (fmtOgre ? 1 : 2)
			callRuby('browse', outputDir + ',' + outputFile + ',' + fmtVal)
		}
		function indexOfExt(str, ext) {
			idx = str.lastIndexOf(ext)
			if (idx == -1)
				return -1;
			if (idx + ext.length == str.length)
				return idx;
			return -1;
		}
		function formatChange() {
			editor = document.getElementById('outFile')
			fmtX = document.getElementById('asXFile').checked
			fmtOgre = document.getElementById('asOgreFile').checked
			exts = ['.x', '.mesh.xml', '.obj']
			fmtVal = fmtX ? 0 : (fmtOgre ? 1 : 2)
			outputFile = editor.value
			
			// remove any known extensions
			for (var i = 0; i < exts.length; ++i)
			{
				idx = indexOfExt(outputFile, exts[i])
				if (idx != -1)
				{
					outputFile = outputFile.substr(0, idx)
					break;
				}
			}
				
			outputFile = outputFile + exts[fmtVal]
			editor.value = outputFile

			document.getElementById('fixTexsZ').disabled = (fmtVal != 2)
		}
		function uvHelperChange() {
			uv = document.getElementById('useUVHelper').checked
			//document.getElementById('fixUVZ').disabled = !uv
		}
		function exportFunc() {
			outputDir = document.getElementById('outDir').value
			outputFile = document.getElementById('outFile').value
			smart = document.getElementById('smartFaces').checked
			front = document.getElementById('frontFaces').checked
			back = document.getElementById('backFaces').checked
			uv = document.getElementById('useUVHelper').checked
			sel = document.getElementById('exportSelected').checked
			tex = document.getElementById('exportTextured').checked
			fmtVal = document.getElementById('asXFile').checked ? 0 : (document.getElementById('asOgreFile').checked ? 1 : 2)
			fixTexs = document.getElementById('fixTexsZ').checked
			fixUV = document.getElementById('fixUVZ').checked
			callRuby('export', outputDir + ',' + outputFile + ',' + smart + ',' + front + ',' + back + ',' + uv + ',' + sel + ',' + tex + ',' + fmtVal + ',' + fixTexs + ',' + fixUV)
		}
		</script>
	</head>
	<body>
		<input type='button' onclick='exportFunc()' value='Export'><br>
		<div>
		<div id='progress' style='color: red'>
		</div>
		Output directory:<br>
		<input id='outDir' type='text' value='#{dir}' size='60'><br>
		Output file: 
		<input type='button' onclick='browseFunc()' value='Browse'><br>
		<input id='outFile' type='text' value='#{name}' size='60'><br>
		<hr>
		<input id='asXFile' type='radio' onclick='formatChange()' name='xfile' value='X File'> Export as DirectX .X file<br>
		<input id='asOgreFile' type='radio' onclick='formatChange()' name='xfile' value='Ogre File'> Export as Ogre3d .MESH.XML file<br>
		<input id='asObjFile' type='radio' onclick='formatChange()' name='xfile' value='OBJ File' checked='checked'> Export as Wavefront .OBJ file<br>
		<p style='padding-left: 20pt'>
		<input id='fixTexsZ' type='checkbox' value='Fix Textures'> alternative texture export (use when textures are missing)
		</p>
		<hr>
		<input id='smartFaces' type='checkbox' value='Smart faces' checked='checked'> Guess which faces to export<br>
		<input id='frontFaces' type='checkbox' value='Front faces'> Export front faces<br>
		<input id='backFaces' type='checkbox' value='Back faces'> Export back faces<br>
		<input id='useUVHelper' type='checkbox' onclick='uvHelperChange()' value='Use UVHelper' checked='checked'> Use UVHelper (try toggling if textures are messed up)<br>
		<input id='fixUVZ' type='checkbox' value='Use Simple UV'> Use alternative texture coordinates (try toggling if textures are messed up)<br>
		<input id='exportSelected' type='checkbox' value='Export Selected'> Export selected only<br>
		<input id='exportTextured' type='checkbox' value='Export Textured'> Export textured only<br>
		<hr>
		<input type='button' onclick='rubyFunc()' value='Show Ruby Console'> (use for debugging)<br>
		<hr>
		<div id='console' style='color: grey; font-family: monospace; font-size: smaller'>
		</div>
		</div>
	</body>
	</html>
	"
	
	#print html
	my_dialog.set_html(html)
	
	my_dialog.show {
  		#my_dialog.execute_script(
  		#	"document.getElementById('ala').innerHTML = '<b>Hi There!</b>'");
	}
end

#-----------------------------------------------------------------------------

def exportXFileEngine(entities, trans, tw,
						materials, indexer, points, uvs, normals, faces,
						smartInfo, print_callback, fmtVal, fixUV)

  entities.each { |ent|
    next if ent.hidden? or not ent.layer.visible?

    case ent.typename
		when "ComponentInstance"
			print_callback.call(true, "Exporting Component: " + ent.definition.name + " " + ent.name)
		    exportXFileEngine(ent.definition.entities, trans*ent.transformation, tw,
							materials, indexer, points, uvs, normals, faces,
							smartInfo, print_callback, fmtVal, fixUV)
			#print_callback.call(true, "Done.")

		when "Group"
			print_callback.call(true, "Exporting Group: " + ent.name)
		    exportXFileEngine(ent.entities, trans*ent.transformation, tw,
							materials, indexer, points, uvs, normals, faces,
							smartInfo, print_callback, fmtVal, fixUV)
			#print_callback.call(true, "Done.")
    end
  }

	ss = entities
	ss = ss.select {|ent| 
		(ent.kind_of? Sketchup::Face) && !ent.hidden? && ent.layer.visible?
	}
	if ss.empty?
		print_callback.call(true, "Nothing to export.")
		return false
	end

	

	smarty = {}
	ss = ss.select {|face|
		matF = nil;
		matF = face.material
		texF = nil
		texF = matF.texture if matF

		matB = nil;
		matB = face.back_material
		texB = nil
		texB = matB.texture if matB

		smart = 0
		smart |= 1 if smartInfo["frontFaces"]
		smart |= 2 if smartInfo["backFaces"]
		
		if (smartInfo["smartFaces"])
			smart |= 1 if texF
			smart |= 2 if texB
			if smart == 0
				smart |= 1 if matF
				smart |= 2 if matB
				smart |= 3 if smart == 0
			end
		end
		
		smart &= ~1 if smartInfo["texturedOnly"] && !texF
		smart &= ~2 if smartInfo["texturedOnly"] && !texB
		
		smarty[face] = smart if smart != 0
		(smart != 0)
	}
	if ss.empty?
		print_callback.call(true, "Nothing to export: no faces chosen")
		return false
	end

	if smartInfo["textureWritePass"]
		ss.each {|face|
			[true, false].each {|front|
				if (smarty[face] & (front ? 1 : 2)) != 0
					tw.load(face, front)
				end
			}
		}
		return true
	end

	# Create transformation w/out translation for normals
	narray=trans.to_a
	narray[12..16]=[0,0,0,1]
	ntrans = Geom::Transformation.new(narray)
	
	ss.each {|face|

	  [true, false].each { |front|
		
		#print_callback.call(false, ".")
		if (smarty[face] & (front ? 1 : 2)) == 0
			next
		end
		
		mat = nil;
		mat = face.material if front
		mat = face.back_material if !front
		tex = nil
		tex = mat.texture if mat
	

		texFile = nil
		if tex
			handle = tw.handle(face, front)
			f.puts "Texture handle invalid!" if handle < 0
			texFile = tw.filename(handle).to_s
		end

	
		mname = "Default_Material"
		if mat
			mname = "_" + mat.name
			mname = mname + "-" + texFile if texFile
			mname = mname.gsub(/[^a-zA-Z0-9]/, "_")
			alpha = 255
			alpha = mat.alpha if mat.use_alpha?
			materials[mname] = [mname, mat.color, texFile, alpha]
			#puts(materials[mname])
		else
			materials[mname] = [mname, nil, nil, nil]
		end

		if tex
			uvHelp = face.get_UVHelper(true, true, tw)

			#retval = tw.write(face, front, fname + tw.filename(handle).to_s)
			#retval = (retval == FILE_WRITE_OK)
	        
			mus = []
			mvs = []
			face.outer_loop.vertices.each { |vertex|
				pos = vertex.position
				if front
					uvq = uvHelp.get_front_UVQ(pos).to_a
				else
					uvq = uvHelp.get_back_UVQ(pos).to_a
				end
			    #mus << (u.x/u.z)
			    #mvs << (u.y/u.z)
			    mus << (uvq.x)
			    mvs << (uvq.y)
			    puts "#{uvq.x} #{uvq.y} #{uvq.z}" if !(1.0-0.0001 < uvq.z && uvq.z < 1.0+0.0001)
			}
			minu = mus.min
			minv = mvs.min
			maxu = mus.max
			maxv = mvs.max
			#print("#{minu} #{minv} #{maxu} #{maxv}\n")
		end

		#print_callback.call(true, "Debug kill.")
		#next
	  
		mesh = face.mesh 7

		#puts("Mesh:")
		for p in (1..mesh.count_polygons)
			poly = mesh.polygon_points_at(p)
			ipoly = mesh.polygons[p - 1]
			#print "Poly #{poly}\n"

			theFace = []
			
			for i in 0..(poly.size - 1)
				idx = i
				idx = (poly.size - 1) - i if !front
			
				pos = poly[idx]
				
				uq = 0.0
				vq = 0.0
				if tex
					if smartInfo["useUVHelper"]
						if front
							uvq = uvHelp.get_front_UVQ(pos).to_a
						else
							uvq = uvHelp.get_back_UVQ(pos).to_a
						end
						#uq = ((uvq.x/uvq.z) - minu) / (maxu - minu) #(mat.texture.width)
						#vq = ((uvq.y/uvq.z) - minv) / (maxv - minv) #(mat.texture.height)
						if fixUV
							uq = (uvq.x)
							vq = (uvq.y)
						else
							uq = ((uvq.x) - minu) #/ (maxu - minu)
							vq = -((uvq.y) - minv) #/ (maxv - minv)
						end
						#printf("%8.4f %8.4f %8.4f %8.4f %8.4f %8.4f\n", uq, vq, (maxu - minu), (maxv - minv), mat.texture.width, mat.texture.height)
						#printf("%8.4f %8.4f %8.4f\n", uvq.x, uvq.y, uvq.z)
						#uq *= mat.texture.width / (maxu - minu)
						#vq *= mat.texture.height / (maxv - minv)
					else
						ni = ipoly[idx]
						ni = - ni if ni < 0
						uvss = mesh.uv_at(ni, 1)
						if fixUV
							uq = uvss.x
							vq = uvss.y
						else
							uq = uvss.x / mat.texture.width
							vq = uvss.y / mat.texture.height
						end
						#printf("%8.4f %8.4f %8.4f\n", uvss.x, uvss.y, uvss.z)
					end
				end

				ni = ipoly[idx]
				ni = - ni if ni < 0
				norm = mesh.normal_at(ni)
				n = norm
				n = n.reverse if (!front)
				n = ntrans * n
				n = n.normalize

				pt, n, uv = out_point(fmtVal, trans * pos, n, uq, vq)
				
				pidx = 0
				uidx = 0
				if fmtVal == FMT_X
					ptuv = pt + uv
					pidx = indexer.index(ptuv)
					if !pidx
						pidx = indexer.size
						points.push(pt) 
						uvs.push(uv) 
						indexer.push(ptuv) 
					end
					uidx = pidx
				else
					pidx = points.index(pt)
					if !pidx
						pidx = points.size
						points.push(pt) 
					end

					uidx = uvs.index(uv)
					if !uidx
						uidx = uvs.size
						uvs.push(uv) 
					end
				end
				
				nidx = normals.index(n)
				if !nidx
					nidx = normals.size
					normals.push(n) 
				end

				theFace.push([pidx, uidx, nidx])
			end

			#puts "TheFace #{theFace}"
			faces.push([theFace, mname])
		end
	  }
	}
	#print_callback.call(false, "\n")
	
	return true
end

#-----------------------------------------------------------------------------

def exportXFile(fname, outDir, fmtVal, fixTexs, fixUV, smartFaces, export_front_faces, export_back_faces,
				use_uv_helper, selectedOnly, texturedOnly, print_callback)

	if !fname || !outDir
		print_callback.call(true, "Empty file or directory name.")
		return false
	end

	model = Sketchup.active_model
	ss = model.active_entities
	ss = model.selection if selectedOnly

	smartInfo = {}
	smartInfo["smartFaces"] = smartFaces
	smartInfo["frontFaces"] = export_front_faces
	smartInfo["backFaces"] = export_back_faces
	smartInfo["useUVHelper"] = use_uv_helper
	smartInfo["selectedOnly"] = selectedOnly
	smartInfo["texturedOnly"] = texturedOnly

	print_callback.call(true, "Exporting textures to: " + outDir)
    tw = Sketchup.create_texture_writer
	smartInfo["textureWritePass"] = true
	
	    exportXFileEngine(ss, Geom::Transformation.new, tw,
						nil, nil, nil, nil, nil, nil,
						smartInfo, print_callback, fmtVal, fixUV)

    result = tw.write_all(outDir, false)
	print_callback.call(true, "Writing textures failed!") if !result
	
	print_callback.call(true, "Analyzing geometry")
	
	smartInfo["textureWritePass"] = false
	materials = {}
	indexer = []
	points = []
	uvs = []
	normals = []
	faces = []
	
		exportXFileEngine(ss, Geom::Transformation.new, tw,
						materials, indexer, points, uvs, normals, faces,
						smartInfo, print_callback, fmtVal, fixUV)

	print_callback.call(true, "Exporting geometry to: " + fname)
	f = File.new(fname, "w")
	
	if fmtVal == FMT_X
		f.puts(directx_header())
		out_materials(fmtVal, fixTexs, f, materials)
		f.puts "Mesh mesh_0{"
		out_tab(fmtVal, f, points)
		out_faces(fmtVal, f, faces)
		f.puts "  MeshMaterialList {"
		out_face_materials(fmtVal, f, faces, materials)
		f.puts "  }"
		f.puts "  MeshTextureCoords {"
		out_tab(fmtVal, f, uvs, "", "")
		f.puts "  }"
		f.puts "  MeshNormals {"
		out_tab(fmtVal, f, normals)
		out_normals(fmtVal, f, faces)
		f.puts "  }"
		f.puts "}"
	elsif fmtVal == FMT_OGRE
		f.puts(ogre_header())

		submeshes = help_group_by(faces) {|fc| fc[1]} # (mat -> faces) ...
		# we'll sort submeshes alphabetically by  material name
		submeshes = submeshes.sort_by { |mname, faces| mname } # [mat, faces] ...
		
		f.puts "<mesh>
	<submeshes>
"
		submeshes.each {|mname, faces|
			f.puts "		<submesh material=\"#{mname}\" usesharedvertices=\"false\" use32bitindexes=\"false\" operationtype=\"triangle_list\">"
			out_faces(fmtVal, f, faces, "			")
			f.puts "			<geometry vertexcount=\"#{points.size}\">
				<vertexbuffer positions=\"true\" normals=\"true\" texture_coord_dimensions_0=\"float2\" texture_coords=\"1\">
"
			out_tab(fmtVal, f, points, "\n", "", "					")
			f.puts "				</vertexbuffer>"
			f.puts "			</geometry>"
			f.puts "		</submesh>"
		}

		f.puts "	</submeshes>
	<submeshnames>"
		
		# each_with_index is not available in Ruby 1.8.6 used in SketchUp 8
		(0..submeshes.size-1).zip(submeshes).each {|i, mnamefaces|
			f.puts "		<submeshname name=\"submesh_#{mnamefaces[0]}\" index=\"#{i}\" />"
		}
		f.puts "	</submeshnames>
</mesh>"
		
		#write material file here

		mtlName = fname
		idx = mtlName.downcase.rindex(".mesh.xml")
		if idx > -1
			mtlName = mtlName[0, idx]
		end
		mtlName = mtlName + ".material"
		
		print_callback.call(true, "Exporting materials to: " + mtlName)
		fm = File.new(mtlName, "w")
		fm.puts(ogre_header_mat())
		fm.puts "\n"
		out_materials(fmtVal, fixTexs, fm, materials)
		fm.close;
	else
		f.puts(obj_header())
		f.puts "\n"
		name, dir, slash = processFileName(fname, FMT_OBJ)
		name[-4..-1] = ".mtl"
		f.puts("mtllib " + name)
		f.puts "\n"
		out_tab(fmtVal, f, points, "\n")
		f.puts "\n"
		out_tab(fmtVal, f, uvs, "\n", "")
		f.puts "\n"
		out_tab(fmtVal, f, normals, "\n")
		f.puts "\n"
		out_faces(fmtVal, f, faces)
		f.puts "\n"

		#write material file here

		mtlName = fname
		idx = mtlName.downcase.rindex(".obj")
		if idx > -1
			mtlName = mtlName[0, idx]
		end
		mtlName = mtlName + ".mtl"
		
		print_callback.call(true, "Exporting materials to: " + mtlName)
		print_callback.call(true, "WARNING: .OBJ format does not support spaces in texture filenames!")
		fm = File.new(mtlName, "w")
		fm.puts(obj_header())
		fm.puts "\n"
		out_materials(fmtVal, fixTexs, fm, materials)
		fm.close;
	end

	f.close	
	print_callback.call(true, "Done.")
	
	return true
end

#-----------------------------------------------------------------------------

end


if( not file_loaded?("ZbylsXExporter.rb") )
    #add_separator_to_menu("Plugins")
    plugins_menu = UI.menu("Plugins")
    plugins_menu.add_item("Zbyl's X Exporter...") { ZbylsXExporter.exportXFileUI }
    #plugins_menu.add_item("Zbyl's X Exporter Reload") { ZbylsXExporter.reload }
end

file_loaded("ZbylsXExporter.rb")

#-----------------------------------------------------------------------------

